% demo_denoising - image denoising experiments in which the image is reordered using patch-based
% permutations and then filtered using learned filters.
% demo_denoising generates the image denoising results reported in Table III in:
% I. Ram, M. Elad, and I. Cohen, "Image Processing using Smooth Ordering of its Patches", 
% IEEE Trans. Image Processing, Vol. 22, Issue 7, July 2013, pp. 2764-2774.
% =====================================================================================
% Idan Ram
% Department of Electrical Engineering
% Technion, Haifa 32000 Israel
% idanram@tx.technion.ac.il
%
% November 2013
% =====================================================================================

clear
close all
clc

%%%% choosing the test image %%%%

picNumber = 1; % choose what image to denoise out of 6 images
if picNumber == 1
    picName = 'house.png';
elseif picNumber == 2 
    picName = 'Lena512.png';
elseif picNumber == 3
    picName = 'barbara.png';
elseif picNumber == 4    
    picName = 'boat.png';
elseif picNumber == 5    
    picName = 'fingerprint.png';
elseif picNumber == 6    
    picName = 'peppers256.png';
end

%%%% parameters %%%%
%%%% N and B - number of rows in a square patch and square search area, used in the first iteration.
%%%% N2 and B2 - number of rows in a square patch and square search area, used in the second iteration.
%%%% C and eps1 - first iteration design paramters
%%%% C2 and eps2 - second iteration design paramters

sigma_n = 5; % choose sigmn_n value out of [ 5 , 10 , 15 , 20 , 25 , 50 , 75 , 100 ]
trainFlag = 0; % set to 1 in order to calculate the filters
permNum = 10; % number of calculated permutations
if sigma_n == 5
    N = 5; C = 2.2; % first iteration parameters
    N2 = 4; C2 = 1.2; % second iteration parameters
elseif sigma_n == 10
    N = 6; C = 1.6; % first iteration parameters
    N2 = 4; C2 = 0.8; % second iteration parameters
elseif sigma_n == 15
    N = 7; C = 1.4; % first iteration parameters
    N2 = 4; C2 = 0.6; % second iteration parameters
elseif sigma_n == 20
    N = 8; C = 1.3; % first iteration parameters
    N2 = 4; C2 = 0.5; % second iteration parameters
elseif sigma_n == 25
    N = 8; C = 1.2; % first iteration parameters
    N2 = 4; C2 = 0.4; % second iteration parameters
elseif sigma_n == 50
    N = 14; C = 1.1; % first iteration parameters
    N2 = 5; C2 = 0.3; % second iteration parameters
elseif sigma_n == 75
    N = 16; C = 1.1; % first iteration parameters
    N2 = 6; C2 = 0.2; % second iteration parameters
elseif sigma_n == 100
    N = 16; C = 1.1; % first iteration parameters
    N2 = 8; C2 = 0.1; % second iteration parameters
end
B = 61; eps1 = 10.^6; hLength = 25; % first iteration parameters
B2 = 361; eps2 = 10.^3; hLength2 = 25; % second iteration parameters

%%%% reading training and test images %%%%

picData = cell( 4 ,3 );

%%%% train images %%%%

picData{ 1 , 1 } = '../images/man.png';
picData{ 2 , 1 } = '../images/couple.png';
picData{ 3 , 1 } = '../images/hill.png';
picData{ 4 , 1 } = [ '../images/' , picName ];

rand('seed', 0);
randn('seed', 0);

%%%% creating noisy images %%%%

for m = 1 : 4
    origPic = double( imread( picData{ m , 1 } ) );
    [ N_m , N_n ] = size( origPic );
    noisyPic = origPic + sigma_n * randn( N_m , N_n );
    picData{ m , 2 } = origPic;
    picData{ m , 3 } = noisyPic;
end
mseIn = mean( mean( ( picData{ 4 , 2 } - picData{ 4 , 3 } ).^2 ) );
PSNRIn = 20*log10( 255 ./ sqrt( mseIn ) );
figure(2)
subplot( 221 ); imshow( picData{ 4 , 2 } , [ 0 , 255 ] , 'border' , 'tight' ); title( 'Original image' );
subplot( 222 ); imshow( picData{ 4 , 3 } , [ 0 , 255 ] , 'border' , 'tight' ); title( 'Noisy image' ); xlabel(['PSNR=' , num2str( PSNRIn )])
drawnow

%%%% Iteration 1 %%%%

%%%% learning filters %%%%

if trainFlag == 1
    disp('iteration 1 - learning filters')
    XTXMat = zeros( 2 * hLength , 2 * hLength , 1 , 3 );
    XTyMat = zeros( 2 * hLength , 1 , 3 );
    for m = 1:3
        disp( [ 'learning from image ' , num2str( m ) ' / 3' ] )
        [ XTXMat( : , : , : , m ) , XTyMat( : , : , m ) ] =  filterDenoise( picData{ m , 3 } , picData{ m , 2 } , permNum , sigma_n , N , B , hLength , eps1 , C );
    end
    XTXMatTotal = sum( XTXMat , 4 );
    XTyMatTotal = sum( XTyMat , 3 );
    hVec =  inv( XTXMatTotal ) * XTyMatTotal; % learned filters
%     eval( [ 'save trainedFilters1/trainedFiltersSigma_n' , num2str( sigma_n ) , ' hVec' ] );
else
    eval( [ 'load trainedFilters1/trainedFiltersSigma_n' , num2str( sigma_n ) ] );
end
figure(1)
subplot(221); plot( hVec( 1 : hLength ) ); xlabel('Samples'); ylabel('Amplitude'); title('h_s - first iteration')
subplot(222); plot( hVec( hLength + 1:2*hLength ) ); xlabel('Samples'); ylabel('Amplitude'); title('h_e - first iteration')
drawnow

%%%% denoising images %%%%

disp('iteration 1 - denoising image')
disp( 'displaying the average of the denoised images obtained with several permutations' )
disp( '( i.e. using cycle spinning )' )
recPic =  filterDenoise( picData{ 4 , 3 } , picData{ 4 , 2 } , permNum , sigma_n , N , B , hLength , eps1 , C , hVec );
mseOut = mean( mean( ( recPic - picData{ 4 , 2 } ).^2 ) );
PSNROut = 20*log10( 255 ./ sqrt( mseOut ) );
figure( 2 )
subplot( 223 ); imshow( recPic , [ 0 , 255 ] , 'border' , 'tight' ); title( '1 iteration result' ); xlabel( [ 'PSNR=' , num2str( PSNROut ) ] )
drawnow

%%%% Iteration 2 %%%%

%%%% learning filters %%%%

rand('seed', 0);
randn('seed', 0);

if trainFlag == 1
    disp('iteration 2 - learning filters , NOTICE: THIS MAY TAKE SOME TIME TO RUN')
    XTXMat = zeros( 2 * hLength2 , 2 * hLength2 , 1 , 3 );
    XTyMat = zeros( 2 * hLength2 , 1 , 3 );
    for m = 1:3
        disp( [ 'learning from image ' , num2str( m ) ' / 3' ] )
        tempRecPic =  filterDenoise( picData{ m , 3 } , picData{ m , 2 } , permNum , sigma_n , N , B , hLength , eps1 , C , hVec );
        [ XTXMat( : , : , : , m ) , XTyMat( : , : , m ) ] =  ...
            filterDenoise( picData{ m , 3 } , picData{ m , 2 } , permNum , sigma_n , N , B , hLength , eps1 , C , hVec , ...
            tempRecPic , N2 , B2 , hLength2 , eps2 , C2 );
    end
    XTXMatTotal = sum( XTXMat , 4 );
    XTyMatTotal = sum( XTyMat , 3 );
    hVec2 =  inv( XTXMatTotal ) * XTyMatTotal;
%     eval( [ 'save trainedFilters2/trainedFilters2Sigma_n' , num2str( sigma_n ) , ' hVec2' ] );
else
    eval( [ 'load trainedFilters2/trainedFilters2Sigma_n' , num2str( sigma_n ) ] );
end
figure(1)
subplot(223); plot( hVec2( 1 : hLength ) ); xlabel('Samples'); ylabel('Amplitude'); title('h_s - second iteration')
subplot(224); plot( hVec2( hLength + 1:2*hLength ) ); xlabel('Samples'); ylabel('Amplitude'); title('h_e - second iteration')
drawnow

%%%% denoising images %%%%

disp('iteration 2 - denoising image , NOTICE: THIS MAY TAKE SOME TIME TO RUN')
disp( 'displaying the average of the denoised images obtained with several permutations' )
disp( '( i.e. using cycle spinning )' )
recPic2 =  ...
    filterDenoise( picData{ 4 , 3 } , picData{ 4 , 2 } , permNum , sigma_n , N , B , hLength , eps1 , C , hVec , ...
    recPic , N2 , B2 , hLength2 , eps2 , C2 , hVec2 );
mseOut2 = mean( mean( ( recPic2 - picData{ 4 , 2 } ).^2 ) );
PSNROut2 = 20*log10( 255 ./ sqrt( mseOut2 ) );
figure( 2 )
subplot( 224 ); imshow( recPic2 , [ 0 , 255 ] , 'border' , 'tight' ); title( '2 iterations result' ); xlabel( [ 'PSNR=' , num2str( PSNROut2 ) ] )