% filtDenoiseTrain - if the learned filters vector hVec is unknown,
% calculates the matrices XTXMat and XTyMat used to obtain it.
% Otherwise obtain a denoised image by filtering the reordered input image noisyPic with hVec.
%
% [ out , out2 ] =  filtDenoiseTrain( noisyPic , origPic , N , permNum , B , hLength , randVec , randVec2 , eps1 , sigma_n , C ) 
% calculates the matrices XTXMat and XTyMat from the image noisyPic.
% In this case: out = XTXMat and out2 = XTyMat.
%
% out =  filtDenoiseTrain( noisyPic , origPic , N , permNum , B , hLength , randVec , randVec2 , eps1 , sigma_n , C , hVec ) 
% reorder the image noisyPic using patches from this image, and uses the filters vector hVec to filter the result. 
% In this case: out = recPic.
%
% [ out , out2 ] =  filtDenoiseTrain( noisyPic , origPic , N , permNum , B , hLength , randVec , randVec2 , eps1 , sigma_n , C , [] , recPicOld ) 
% calculates the matrices XTXMat and XTyMat from the image recPicOld.
% In this case: out = XTXMat and out2 = XTyMat.
%
% out =  filtDenoiseTrain( noisyPic , origPic , N , permNum , B , hLength , randVec , randVec2 , eps1 , sigma_n , C , hVec , recPicOld ) 
% reorder the image noisyPic using patches from the image recPicOld, and uses the filters vector hVec to filter the result. 
% In this case: out = recPic.
% =====================================================================================
% Input:
% noisyPic, origPic - the noisy anf clean input images, respectively.
% N, B - number of rows in a square patch and square search area.
% permNum - number of used permutations.
% hLength - the length of each the learned filters.
% randVec , randVec2 - random variables used to calculate the permutations.
% eps1 , C - design paramters.
% hVec - filters vector used to filter the reordered noisy image.
% sigma_n - noise standard deviation.
% recPicOld - initialization image .
% =====================================================================================
% Output:
% out - contains either the matrix XTXMat or the denoised image recPic.
% out2 - contains the matrix XTyMat.
% =====================================================================================
% Idan Ram
% Department of Electrical Engineering
% Technion, Haifa 32000 Israel
% idanram@tx.technion.ac.il
%
% November 2013
% =====================================================================================

function [ out , out2 ] =  filtDenoiseTrain( noisyPic , origPic , N , permNum , B , hLength , randVec , randVec2 , eps1 , sigma_n , C , hVec , recPicOld  )

[ N_m , N_n ] = size( noisyPic );
FMat = im2col( noisyPic , [ N , N ] , 'sliding' );
Weight = reshape( mxSubim2im( ones( N^2 * ( N_m - N + 1 )*( N_n - N + 1 ) , 1 ) , N , N_m ) , N_m , N_n );

% extract patches

if nargin > 12
    FMatRec = im2col( recPicOld , [ N , N ] , 'sliding' );
    X_l = FMatRec;
else
    X_l = FMat;
end
ixPic = reshape( 1 : ( N_m - N + 1 ) * ( N_n - N + 1 ) , [ N_m , N_n ] - N + 1 );
tempIx = ixPic(:);
[ cur_rows , cur_cols ] = ind2sub( [ N_m , N_n ] - N + 1 , tempIx );
ixMat = [ cur_rows , cur_cols ];
X_lLength = size( X_l , 2 );
x_restMat = [ X_l ; ones( 1 , X_lLength ) ; -1/2 * sum( X_l.^2 ) ];
x_1Mat = [ X_l ; -1/2 * sum( X_l.^2 ) ; ones( 1 , X_lLength ) ];

% classify patches into smoooth patches and patches with texture or edges

if nargin > 12
    tempIx = std( FMatRec )' < C * sigma_n;
else
    tempIx = std( FMat )' < C * sigma_n;
end
X_lLength1 = sum( tempIx );
X_lLength2 = sum( ~tempIx );

tempPicMat2 = zeros( N_m * N_n , 2 * hLength );
for k = 1 : permNum
    
    disp( [ 'calculating and applying permutation ' , num2str( k ) ' / ' , num2str( permNum ) ] )
    
    % find a permutation from smooth patches
    
    randStartIx1 = ceil( X_lLength1 * randVec( k , 1 ) );
    tempIx1 = 1 : size( FMat , 2 );
    tempIx1 = tempIx1( tempIx )';
    randStartIx = tempIx1( randStartIx1 );
    freePoints = int32( ~tempIx );
    sortedIx1 = mxCalcPermRand2Sets( x_1Mat( : ) , x_restMat( : ) , ixMat( : ) , freePoints , N^2 + 2 , ...
        X_lLength , X_lLength1 , N_m - N + 1 , randStartIx , B , randVec2( 1:X_lLength1 , k , 1 ) , eps1 );
    sortedIx1 = double( sortedIx1 );
    
    % find a permutation from patches with texture or edges
    
    randStartIx2 = ceil( X_lLength2 * randVec( k , 2 ) );
    tempIx2 = 1 : size( FMat , 2 );
    tempIx2 = tempIx2( ~tempIx )';
    randStartIx = tempIx2( randStartIx2 );
    freePoints = int32( tempIx );
    sortedIx2 = mxCalcPermRand2Sets( x_1Mat( : ) , x_restMat( : ) , ixMat( : ) , freePoints , N^2 + 2 , ...
        X_lLength , X_lLength2 , N_m - N + 1 , randStartIx , B , randVec2( 1:X_lLength2 , k , 2 ) , eps1 );
    sortedIx2 = double( sortedIx2 );
    
    tempPicMat = zeros( N_m * N_n , 2 * hLength );
    sortedIxPad1 = [ sortedIx1( floor( hLength / 2 ) : - 1 : 1 ) ;  sortedIx1 ; sortedIx1( end : -1 : end - floor( hLength / 2 ) + 1 ) ];
    sortedFMatPad1 = FMat( : , sortedIxPad1 )';
    sortedIxPad2 = [ sortedIx2( floor( hLength / 2 ) : - 1 : 1 ) ;  sortedIx2 ; sortedIx2( end : -1 : end - floor( hLength / 2 ) + 1 ) ];
    sortedFMatPad2 = FMat( : , sortedIxPad2 )';
    z = hLength;
    for t = 1 : hLength
        blocks = zeros( X_lLength , N^2 );
        blocks2 = zeros( X_lLength , N^2 );
        blocks( sortedIx1 , : )  = sortedFMatPad1( z : z + length( sortedIx1 ) - 1 , : );
        blocks2( sortedIx2 , : )  = sortedFMatPad2( z : z + length( sortedIx2 ) - 1 , : );
        tempPic = reshape( mxSubim2im( blocks(:) , N , N_m ) , N_m , N_n ) ./ Weight;
        tempPic2 = reshape( mxSubim2im( blocks2(:) , N , N_m ) , N_m , N_n ) ./ Weight;
        
        tempPicMat( : , t ) = tempPic( : );
        tempPicMat( : , t + hLength  ) = tempPic2( : );
        z = z - 1;
    end
    tempPicMat2 = tempPicMat2 + tempPicMat;
    if ( nargin > 11 ) && ~isempty( hVec )
        tempRecPic = reshape( tempPicMat2 / k * hVec , N_m , N_n );
        figure(3)
        imshow( tempRecPic , [ 0 , 255 ] )
        tempPSNR = 20*log10( 255 / sqrt( mean( mean( ( tempRecPic - origPic ).^2 ) ) ) );
        title([ 'Denoising result obtained with ' , num2str( k ) , ' permutations - PSNR = ' , num2str( tempPSNR ) ] )
        drawnow
    end
    
end
if ( nargin < 12 ) || isempty( hVec ) % find the matrices XTXMat and XTyMat
    XTXMat = zeros( 2 * hLength , 2 * hLength );
    XTyMat = zeros( 2 * hLength , 1 );
    tempPicMat3 = tempPicMat2( : , [ ceil( hLength / 2 ) - floor( hLength / 2 ) : ceil( hLength / 2 ) + floor( hLength / 2 )  , ...
        ceil( hLength / 2 ) + hLength - floor( hLength / 2 ) : ceil( hLength / 2 ) + hLength + floor( hLength / 2 ) ] ) / permNum;
    XTXMat( 1 : 2 * hLength , 1 : 2 *hLength ) = tempPicMat3' * tempPicMat3;
    XTyMat( 1 : 2 * hLength ) = tempPicMat3' * origPic(:);
    out = XTXMat;
    out2 = XTyMat;
else
    recPic = tempRecPic; % Denoising result obtained with all the permutations
    out = recPic;
end




