% filterDenoise - if the learned filters vectors hVec and hVec2 are unknown,
% calculates the matrices XTXMat and XTyMat used to obtain them.
% Otherwise obtain a reconstructed image by filtering the reordered input image noisyPic with hVec or hVec2.
%
% [ out , out2 ] =  filterDenoise( noisyPic , origPic , permNum , sigma_n , N , B , hLength , eps1 , C ) 
% calculates the matrices XTXMat and XTyMat from the image noisyPic.
% In this case: out = XTXMat and out2 = XTyMat.
%
% out =  filterDenoise( noisyPic , origPic , permNum , sigma_n , N , B , hLength , eps1 , C )
% reorders the image noisyPic using patches from this image, and uses the filters vector hVec to filter the result. 
% In this case: out = recPic.
%
% [ out , out2 ] =  filterDenoise( noisyPic , origPic , permNum , sigma_n , N , B , hLength , eps1 , C , hVec , ...
%    recPicOld ,  N2 , B2 , hLength2 , eps2 , C2 ) 
% calculates the matrices XTXMat and XTyMat from the image recPicOld.
% In this case: out = XTXMat and out2 = XTyMat.
%
% out =  filterDenoise( noisyPic , origPic , permNum , sigma_n , N , B , hLength , eps1 , C , hVec , ...
%    recPicOld ,  N2 , B2 , hLength2 , eps2 , C2 , hVec2 ) 
% reorder the image noisyPic using patches from the image recPicOld, and uses the filters vector hVec2 to filter the result. 
% In this case: out = recPic2.
% =====================================================================================
% Input:
% noisyPic, origPic - the noisy and clean input images, respectively.
% N, B - number of rows in a square patch and square search area, used in the first iteration.
% N2, B2 - number of rows in a square patch and square search area, used in the second iteration.
% permNum - number of used permutations.
% hLength, hLength2  - the length of each the learned filters in the first and second iterations, respectively.
% randVec , randVec2 - random variables used to calculate the permutations.
% eps1 , C - design paramters used in the first iteration.
% eps2 , C2 - design paramters used in the second iteration.
% hVec, hVec2 - filters vector used to filter the reordered noisy image in the first and second iterations, respectively.
% sigma_n - noise standard deviation.
% recPicOld - initialization image .
% =====================================================================================
% Output:
% out - contains either the matrix XTXMat or the filtered images recPic or recPic2.
% out2 - contains the matrix XTyMat.
% =====================================================================================
% Idan Ram
% Department of Electrical Engineering
% Technion, Haifa 32000 Israel
% idanram@tx.technion.ac.il
%
% November 2013
% =====================================================================================

function [ out , out2 ] =  filterDenoise( noisyPic , origPic , permNum , sigma_n , N , B , hLength , eps1 , C , hVec , ...
    recPicOld ,  N2 , B2 , hLength2 , eps2 , C2 , hVec2 )

[ N_m , N_n ] = size( noisyPic );
randVec = rand( permNum , 2 );
randVec2 = rand( N_m * N_n , permNum , 2 );

if nargin < 11 % calculate permutations from noisyPic
    if nargin < 10 % hVec is unknown
        [ XTXMat , XTyMat ] =  ...
            filtDenoiseTrain( noisyPic , origPic , N , permNum , B , hLength , randVec , randVec2 , eps1 , sigma_n , C  );
        out = XTXMat;
        out2 = XTyMat;
    else % use hVec to filter the reordered noisyPic
        recPic =  ...
            filtDenoiseTrain( noisyPic , origPic , N , permNum , B , hLength , randVec , randVec2 , eps1 , sigma_n , C , hVec  );
        out = recPic;
    end
else % calculate permutations from recPicOld
    if nargin < 17 % hVec2 is unknown
        [ XTXMat , XTyMat ] =  ...
            filtDenoiseTrain( noisyPic , origPic , N2 , permNum , B2 , hLength2 , randVec , randVec2 , eps2 , sigma_n , C2 , [] , recPicOld );
        out = XTXMat;
        out2 = XTyMat;
    else % use hVec2 to filter the reordered noisyPic
        recPic2 =  ...
            filtDenoiseTrain( noisyPic , origPic , N2 , permNum , B2 , hLength2 , randVec , randVec2 , eps2 , sigma_n , C2 , hVec2 , recPicOld );
        out = recPic2;
    end
end


