% demo_inpainting - image inpainting experiments in which the image is reordered using patch-based
% permutations and then the missing pixels are reconstructed using cubic interpolation.
% demo_inpainting generates the image inpainting results reported in Table V in:
% I. Ram, M. Elad, and I. Cohen, "Image Processing using Smooth Ordering of its Patches", 
% IEEE Trans. Image Processing, Vol. 22, Issue 7, July 2013, pp. 2764-2774.
% =====================================================================================
% Idan Ram
% Department of Electrical Engineering
% Technion, Haifa 32000 Israel
% idanram@tx.technion.ac.il
%
% November 2013
% =====================================================================================

clear
close all
clc

picNumber = 1; % choose what image to inpaint out of 3 images
if picNumber == 1 
    picName = 'house.png';
elseif picNumber == 2
    picName = 'Lena512.png';
elseif picNumber == 3    
    picName = 'barbara.png';
end

%%%% parameters %%%%

permNum = 10; % number of calculated permutations
N = 16; B = 9; eps1 = 10^2; % first iteration patch size, search area and design parameter
N2 = 8; B2 = 43; eps2 = 10^4; % second iteration patch size, search area and design parameter
N3 = 5; B3 = 55; eps3 = 10^8; % third iteration patch size, search area and design parameter

%%%% reading clean image %%%%

origPic = double( imread( [ '../images/' , picName ] ) );
[ N_m , N_n ] = size( origPic );

%%%% creating corrupted image %%%%

rand('seed', 0);
randn('seed', 0);

mask = ones( size( origPic ) );
permVec = randperm( N_m * N_n );
zerosNum = 0.8;
mask( permVec( 1 : round( N_m * N_n * zerosNum  ) ) ) = 0;
y = mask .* origPic;
mseIn = mean( mean( ( y - origPic ).^2 ) );
PSNRIn = 20*log10( 255 ./ sqrt( mseIn ) );
figure(1)
subplot( 231 ); imshow( origPic , [ 0 , 255 ] , 'border' , 'tight' ); title( 'Original image' );
subplot( 232 ); imshow( y , [ 0 , 255 ] , 'border' , 'tight' ); title( '80% missing pixels' ); xlabel(['PSNR=' , num2str( PSNRIn )]) 
drawnow

%%%% Iteration 1 %%%%

disp( 'iteration 1' )
disp( 'displaying the average of the reconstructed images obtained with several permutations' )
disp( '( i.e. using cycle spinning )' )
randVec = rand( permNum , 1 );
randVec2 = rand( N_m * N_n , permNum );
recPic =  inpaint1DInterp( y , origPic , mask , N , permNum , B , randVec , randVec2( 1:( N_m -  N + 1 )^2 , : ) , eps1  );
mseOut = mean( mean( ( recPic - origPic ).^2 ) );
PSNROut = 20*log10( 255 / sqrt( mseOut ) );
figure(1)
subplot( 234 ); imshow( recPic , [ 0 , 255 ] , 'border' , 'tight' ); title( '1 iteration result' ); xlabel( [ 'PSNR=' , num2str( PSNROut ) ] )
drawnow

%%%% Iteration 2 %%%%

disp( 'iteration 2' )
disp( 'displaying the average of the reconstructed images obtained with several permutations' )
disp( '( i.e. using cycle spinning )' )
randVec = rand( permNum , 1 );
randVec2 = rand( N_m * N_n , permNum );
recPic2 =  inpaint1DInterp( y , origPic , mask , N2 , permNum , B2 , randVec , randVec2( 1:( N_m - N2 + 1 )^2 , : ) , eps2 , recPic  );
mseOut2 = mean( mean( ( recPic2 - origPic ).^2 ) );
PSNROut2 = 20*log10( 255 / sqrt( mseOut2 ) );
figure(1)
subplot( 235 ); imshow( recPic2 , [ 0 , 255 ] , 'border' , 'tight' ); title( '2 iterations result' ); xlabel( [ 'PSNR=' , num2str( PSNROut2 ) ] )
drawnow

%%%% Iteration 3 %%%%

disp( 'iteration 3' )
disp( 'displaying the average of the reconstructed images obtained with several permutations' )
disp( '( i.e. using cycle spinning )' )
randVec = rand( permNum , 1 );
randVec2 = rand( N_m * N_n , permNum );
recPic3 =  inpaint1DInterp( y , origPic , mask , N3 , permNum , B3 , randVec , randVec2( 1:( N_m - N3 + 1 )^2 , : ) , eps3 , recPic2  );
mseOut3 = mean( mean( ( recPic3 - origPic ).^2 ) );
PSNROut3 = 20*log10( 255 / sqrt( mseOut3 ) );
figure(1)
subplot( 236 ); imshow( recPic3 , [ 0 , 255 ] , 'border' , 'tight' ); title( '3 iterations result' ); xlabel( [ 'PSNR=' , num2str( PSNROut3 ) ] )
