% inpaint1DInterp - reconstruct an image from its corrupted version of
% which contains missing pixels by:
%   1. using the patches to calcualte a 1D smoothing permutation.
%   2. applying this permutation to subimages of the corrupted image.
%   3. reconstruct them missing pixels in each subimage using cubic
%   interpolatiion.
%   4. average the the values obtained in the different subimages for each
%   pixel.
%
% recPic =  inpaint1DInterp( y , origPic , mask , N , permNum , BSizeVec , randVec , randVec2 , eps1Vec )
% reorder the corrupted image y into a 1D signal using patches from this image, and uses cubic interpolation to recover the missing pixels.
%
% recPic =  inpaint1DInterp( y , origPic , mask , N , permNum , BSizeVec , randVec , randVec2 , eps1Vec , recPicOld )
% reorder the corrupted image y into a 1D signal using patches from recPicOld, and uses cubic interpolation to recover the missing pixels.
% =====================================================================================
% Input:
% y - the corrupted input image.
% origPic - original image, used for error calculations
% mask - a mask containing the location of the non-missing pixels
% N, B - number of rows in a square patch and square search area.
% permNum - number of used permutations.
% randVec , randVec2 - random variables used to calculate the permutations.
% eps1 - design paramters.
% recPicOld - initialization image .
% =====================================================================================
% Output:
% recPic - reconstructed image.
% =====================================================================================
% Idan Ram
% Department of Electrical Engineering
% Technion, Haifa 32000 Israel
% idanram@tx.technion.ac.il
%
% November 2013
% =====================================================================================

function recPic =  inpaint1DInterp( y , origPic , mask , N , permNum , B , randVec , randVec2 , eps1 , recPicOld  )

[ N_m , N_n ] = size( y );
FMat = im2col( y , [ N , N ] , 'sliding' ); % extract corrupted patches
FMatLength = size( FMat , 2 );
maskMat = im2col( mask , [ N , N ] , 'sliding' );
ixPic = reshape( 1 : ( N_m - N + 1 ) * ( N_n - N + 1 ) , [ N_m , N_n ] - N + 1 );
[ cur_rows , cur_cols ] = ind2sub( [ N_m , N_n ] - N + 1 , ixPic(:) );
ixMat = [ cur_rows , cur_cols ]; % patch coordinates
fRecMat = zeros( N^2 , FMatLength );
recPicMat = zeros( N_m , N_n , permNum );

if nargin < 10
    X_l = FMat;
    X_lLength = size( X_l , 2 );
else
    X_l = im2col( recPicOld , [ N , N ] , 'sliding' );
    X_lLength = size( X_l , 2 );
    x_restMat = [ X_l ; ones( 1 , X_lLength ) ; -1/2 * sum( X_l.^2 ) ];
    x_1Mat = [ X_l ; -1/2 * sum( X_l.^2 ) ; ones( 1 , X_lLength ) ];
end

for k = 1 : permNum
    disp( [ 'calculating and applying permutation ' , num2str( k ) ' / ' , num2str( permNum ) ] )
    randStartIx = ceil( X_lLength * randVec( k ) );
    
    % calculate permutation
    
    if nargin < 10
        sortedIx = double( mxCalcPermInpaintRand( X_l( : ) , maskMat( : ) , ixMat( : ) , N^2 , X_lLength , randStartIx , B , randVec2( : , k ) , eps1 ) );
    else
        sortedIx = double( mxCalcPermRand( x_1Mat( : ) , x_restMat( : ) , ixMat( : ) , N^2 + 2 , X_lLength , N_m - N + 1 , randStartIx , B , randVec2( : , k ) , eps1 , 1 ) );
    end
    
    % perform cubic interpolation
    
    tempIx = 1 : FMatLength;
    tempMask = logical( maskMat( : , sortedIx ) );
    temp_yVec = FMat( : , sortedIx );
    for t = 1 : N^2
        tempIx2 = tempIx( ~tempMask( t , : ) );
        tempIx3 = tempIx( tempMask( t , : ) );
        interp_yVec = temp_yVec( t , : );
        interp_yVec( tempIx2 ) = interp1( tempIx3 ,temp_yVec( t , tempMask( t , : ) ), tempIx2 , 'spline' );
        interp_yVec( interp_yVec < 0 ) = 0;
        interp_yVec( interp_yVec > 255 ) = 255;
        fRecMat( t , : ) = interp_yVec;
    end
    fRecMat( : , sortedIx ) = fRecMat;
    
    %reconsturct image from subimages
    
    blocks = fRecMat;
    IMout = zeros( N_m , N_n );
    Weight = zeros( N_m , N_n );
    [ cols ,  rows ] = meshgrid( 1: N );
    cols = cols( : ); rows = rows( : );
    for i  = 1:length(cols)
        col = cols(i);
        row = rows(i);
        block = reshape( blocks( i , : ) , [ N_m - N + 1 , N_n - N + 1 ] );
        tempMat = isnan( block );
        block( tempMat ) = 0;
        IMout( row : row + N_m - N , col : col + N_m - N ) ...
            = IMout( row : row + N_m - N , col : col + N_m - N ) + block;
        Weight( row : row + N_m - N , col : col + N_m - N ) = ...
            Weight( row : row + N_m - N , col : col + N_m - N ) + double( ~tempMat );
    end
    recPicMat( : , : , k ) = IMout ./ Weight;
    tempRecPicMat = recPicMat( : , : , 1:k );
    tempMat = isnan( tempRecPicMat );
    tempRecPicMat( tempMat ) = 0;
    
    % average the results obtained with the different permutations
    
    tempRecPic = sum( tempRecPicMat , 3 ) ./ sum( double( ~tempMat ) , 3 );
    figure(2)
    imshow( tempRecPic , [ 0 , 255 ] )
    tempPSNR = 20*log10( 255 / sqrt( mean( mean( ( tempRecPic - origPic ).^2 ) ) ) );
    title([ 'Inpainting result obtained with ' , num2str( k ) , ' permutations - PSNR = ' , num2str( tempPSNR ) ] ) 
    drawnow
    
end
recPic = tempRecPic;

