/* mxCalcPermInpaintRand - calculates a random permutation from image patches which contain missing pixels

[ sortedIx , randCount ] = mxCalcPermInpaintRand( xVec , xNZVec , ixVec , N , pointsNum , randStartIx , B , randVec , eps );
calculates a random permutation from the image patches by 
 1. starting from the patch with the index randStartIx    
 2. continue from each patch to its unvisited nearest neighbor in a BxB search area surrounding it
=====================================================================================
Input:
xVec - vector containing the image patches.
xNZVec - vector containing locations of the non-missing pixels in the image patches.
ixVec - a vector containing the patch coordinates
N - number of pixels in a patch + 2.  
pointsNum - number of patches to reorder.
randStartIx - the index of the first point in the nearest neighbor search
B - a single dimension of the search area used for each patch
randVec - random variables used to calculate the permutations.
eps - design paramters.
=====================================================================================
Output:
sortedIx - the calculated permutation
randCount - the number of patches for which the second nearest neighbor was selected
=====================================================================================
Idan Ram
Department of Electrical Engineering
Technion, Haifa 32000 Israel
idanram@tx.technion.ac.il

November 2013
===================================================================================== */

#include "mex.h"
#include "math.h"

#define OK 0
#define ERROR -1
#define NARGIN 9
#define NARGOUT 2

int mxCalcPermInpaintRand(
        /* inputs */
        double * xVec,
        double * xNZVec,
        double * ixVec,
        int N,
        int pointsNum,
        int randStartIx,
        int B,
        double * randVec,
        double eps,
        
        /* outputs */
        int * sortedIx,
        int * randCount
        ) {
    
    int n , m , k , t , z , b_x , b_y , empty , empty2 , NZcount , N_m , B_h , picDim , start_x , finish_x , 
            start_y , finish_y , tempMinIx , tempMinIx2 , tempSpaceDist, tempSpaceMinIx , tempSpaceMinVal , emptyS;
    double tempMinVal , tempDist , tempMinVal2;
    int * freePoints;
    double * tempVec , * tempVec2;
    
    freePoints = ( int * ) malloc( pointsNum * sizeof( int ) );
    if(freePoints==NULL) {
        return ERROR;
    }
    
    tempVec = ( double * ) malloc( N * sizeof( double ) );
    if(tempVec==NULL) {
        return ERROR;
    }
    
    tempVec2 = ( double * ) malloc( N * sizeof( double ) );
    if(tempVec==NULL) {
        return ERROR;
    }
    
    for ( n = 0 ; n < pointsNum ; n++ ) {
        freePoints[n] = 0;
    }
    
    /* finding the nn path */
    
    picDim = (int) sqrt( pointsNum );
    B_h = ( B - 1 ) / 2;
    sortedIx[0] = randStartIx;
    randCount[0] = 0;
    freePoints[ sortedIx[0] - 1 ] = 1;
    for ( k = 0 ; k < pointsNum - 1 ; k++ ) {
        
        for ( n = 0 ; n < N ; n++ ) {
            tempVec[n] = xVec[ ( sortedIx[k] - 1 ) * N + n ];
            tempVec2[n] = xNZVec[ ( sortedIx[k] - 1 ) * N + n ];
        }
        
        t = 0;
        empty = 0;
        empty2 = 0;
        emptyS = 0;
        if ( B < picDim ) {
            n = (int) floor( ( sortedIx[k] - 1 ) / picDim );
            m = ( sortedIx[k] - 1 ) % picDim;
            for ( b_x = -B_h ; b_x <= B_h ; b_x++ ) {
                if ( n + b_x >=0 && n + b_x < picDim  ){
                    for ( b_y = -B_h ; b_y <= B_h ; b_y++ ) {
                        if ( m + b_y >=0 && m + b_y < picDim  ){
                            t = ( n + b_x ) * picDim + m + b_y;
                            if ( freePoints[t] != 1 ){
                                tempDist = 0;
                                NZcount = 0;
                                for ( z = 0 ; z < N ; z++ ) {
                                    if ( ( tempVec2[z] == 1 ) && ( xNZVec[ t * N + z ] == 1 ) ){
                                        tempDist = tempDist + ( tempVec[z] -  xVec[ t * N + z ] ) * ( tempVec[z] -  xVec[ t * N + z ] );
                                        NZcount = NZcount + 1;
                                    }
                                }
                                if ( NZcount > 0 ) {
                                    tempDist = tempDist / NZcount;
                                    if ( empty == 0 ) {
                                        empty = 1;
                                        tempMinVal = tempDist;
                                        tempMinIx = t + 1; /* incremented by 1 for matlab */
                                    }
                                    else if ( tempMinVal > tempDist ){
                                        empty2 = 1;
                                        tempMinIx2 = tempMinIx;
                                        tempMinVal2 = tempMinVal;
                                        tempMinVal = tempDist;
                                        tempMinIx = t + 1; /* incremented by 1 for matlab */
                                    }
                                    else if ( ( tempMinVal2 > tempDist ) | ( empty2 == 0 ) ) {
                                        empty2 = 1;
                                        tempMinVal2 = tempDist;
                                        tempMinIx2 = t + 1; /* incremented by 1 for matlab */
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        
        if ( empty == 0 ) {
            for ( m = 0 ; m < pointsNum ; m++ ) {
                
                if ( freePoints[m] != 1 ) {
                    
                    tempDist = 0;
                    tempSpaceDist = 0;
                    NZcount = 0;
                    for ( n = 0 ; n < N ; n++ ) {
                        if ( ( tempVec2[n] == 1 ) && ( xNZVec[ m * N + n ] == 1 ) ){
                            tempDist = tempDist + ( tempVec[n] -  xVec[ m * N + n ] ) * ( tempVec[n] -  xVec[ m * N + n ] );
                            NZcount = NZcount + 1;
                        }
                    }   
                    if ( NZcount > 0 ) {
                        tempDist = tempDist / NZcount;
                        if ( empty == 0 ) {
                            empty = 1;
                            tempMinVal = tempDist;
                            tempMinIx = m + 1; /* incremented by 1 for matlab */
                        }
                        else if ( tempMinVal > tempDist ){
                            empty2 = 1;
                            tempMinIx2 = tempMinIx;
                            tempMinVal2 = tempMinVal;
                            tempMinVal = tempDist;
                            tempMinIx = m + 1; /* incremented by 1 for matlab */
                        }
                        else if ( ( tempMinVal2 > tempDist ) | ( empty2 == 0 ) ) {
                            empty2 = 1;
                            tempMinVal2 = tempDist;
                            tempMinIx2 = m + 1; /* incremented by 1 for matlab */
                        }
                    }
                    else if ( empty == 0 ) {
                        tempSpaceDist = ( ixVec[sortedIx[k] - 1] - ixVec[m] ) * ( ixVec[sortedIx[k] - 1] - ixVec[m] ) 
                        + ( ixVec[sortedIx[k] - 1 + pointsNum] - ixVec[m + pointsNum] ) * ( ixVec[sortedIx[k] - 1 + pointsNum] - ixVec[m + pointsNum] );
                        if ( emptyS == 0 ) {
                            emptyS = 1;
                            tempSpaceMinVal = tempSpaceDist;
                            tempSpaceMinIx = m + 1; /* incremented by 1 for matlab */
                        }
                        else if ( tempSpaceMinVal > tempSpaceDist ){
                            tempSpaceMinVal = tempSpaceDist;
                            tempSpaceMinIx = m + 1; /* incremented by 1 for matlab */
                        }
                    }
                }
            }
        }
        if ( empty == 1 ) {
            tempMinVal = 1/ ( 1 + exp( ( tempMinVal - tempMinVal2 ) / eps ) );
            if ( ( empty2 == 1 ) && ( k < pointsNum - 2 ) && ( randVec[ k + 1 ] > tempMinVal ) ) {
                sortedIx[k + 1] = tempMinIx2;
                randCount[0] = randCount[0] + 1;
            }
            else {
                sortedIx[k + 1] = tempMinIx;
            }
        }
        else{
            sortedIx[k + 1] = tempSpaceMinIx;
        }
        /*
         * printf( " sortedIx = %d \n " , sortedIx[k] );
         */
        freePoints[sortedIx[k + 1]-1] = 1;
    }
    free(freePoints);
    free(tempVec);
    free(tempVec2);
    
    return OK;
    
}

void mexFunction( int nlhs, mxArray *plhs[],
        int nrhs, const mxArray *prhs[] ) {
    /* inputs */
    double * xVec;
    double * xNZVec;
    double * ixVec;
    int N;
    int pointsNum;
    int randStartIx;
    int B;
    double * randVec;
    double eps;
    
    /* outputs */
    int * sortedIx;
    int * randCount;
    
    unsigned int dimsS[2];
    unsigned int dimsC[2];
    int ret_code;
    
    /* Check for proper number of arguments. */
    if(nrhs != NARGIN) {
        mexErrMsgTxt("12 inputs required.");
    } else if(nlhs > NARGOUT) {
        mexErrMsgTxt("Too many output arguments");
    }
    
    /* prhs - array of inputs */
    xVec = mxGetPr(prhs[0]);
    xNZVec = mxGetPr(prhs[1]);
    ixVec = mxGetPr(prhs[2]);
    N = (int) *mxGetPr(prhs[3]);
    pointsNum = (int) *mxGetPr(prhs[4]);
    randStartIx = (int) *mxGetPr(prhs[5]);
    B = ( int )*mxGetPr(prhs[6]);
    randVec = mxGetPr(prhs[7]);
    eps = *mxGetPr(prhs[8]);
    
    dimsS[0] = pointsNum;
    dimsS[1] = 1;
    dimsC[0] = 1;
    dimsC[1] = 1;
    
    /* plhs - array of outputs */
    plhs[0] = mxCreateNumericArray(2, dimsS, mxINT32_CLASS, mxREAL);
    sortedIx = mxGetData(plhs[0]);
    plhs[1] = mxCreateNumericArray(2, dimsC, mxINT32_CLASS, mxREAL);
    randCount = mxGetData(plhs[1]);
    
    /* call the calculating function */
    ret_code = mxCalcPermInpaintRand( xVec , xNZVec , ixVec , N , pointsNum , randStartIx , B , randVec , eps , sortedIx , randCount );
    if (ret_code==ERROR) {
        mexErrMsgTxt("error occured");
    }
}



